--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- Divider - This is an implementation of a 32-bit serial restoring
--              divider as proposed in Hennessy & Patterson textbook
--              - The implementation waits for a 1 clock cycle initialization
--              input control signal. Then it computes the multiplication in
--              the next XXX clock cycles. The last of these cycles generates
--				an output signal (end_div) with duration of exactly
--				1 clock cycle  to indicate the end of the division,
--				with the result then available at the data outputs.
--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_unsigned.all;

entity div_serial is
	generic(size: integer := 32);
	port
	(
		reset, clock: in std_logic;
		start: in std_logic;
		Rs: in std_logic_vector((size-1) downto 0);
		B_Rt: in std_logic_vector((size-1) downto 0);
		end_div: out std_logic;
		A_Lo, P_Hi: out std_logic_vector((size-1) downto 0)
	);
end div_serial;

architecture div_serial of div_serial is

	type State_type is (inicio, desloca, subtrai, esc_quoc, restaura, fim);
	signal EA, PE: State_type;

	signal regs_PA : std_logic_vector( (size*2) downto 0);
	-- concatenation of registers P e A:
	-- P - auxiliar reg that will contain the remainder at the end
	-- A - initially contains the dividend, at the end contains teh quotient
	signal regB : std_logic_vector( size downto 0);
	-- contains the divisor
	signal soma_sub :    std_logic_vector( size downto 0);
	-- control signal for add/subtract module: 0-add, 1-subtract
	signal cont : integer;

begin

	soma_sub <=	regs_PA(size*2 downto size) - regB when EA = subtrai else
				regs_PA(size*2 downto size) +  regB;

	A_Lo <= regs_PA( size-1 downto 0);
	P_Hi  <= regs_PA( size*2-1 downto size);
	end_div <= '1' when EA=fim else '0';

	-- Process that implements registers regs_PA, regB and cont
	process(clock)
	begin
		if clock'event and clock='1' then
			if EA=inicio then
				regs_PA(size*2 downto size) <= (others=>'0');
				regs_PA(size-1 downto 0) <= Rs;
				regB <= '0'& B_Rt;
				cont <= 0;
			elsif EA=desloca then
				regs_PA(size*2 downto 0) <= regs_PA(((size*2)-1) downto 0) & '0';
			elsif EA=subtrai then
				regs_PA (size*2 downto size) <= soma_sub;
			elsif EA=esc_quoc then
				regs_PA(0) <= not regs_PA(size*2);
			elsif EA=restaura then
				cont  <= cont + 1;
				if regs_PA(size*2)='1' then
					regs_PA (size*2 downto size) <= soma_sub;
				end if;
			end if;
		end if;
	end process;

-------------------------------------------------
-- State machine to control the division process
-------------------------------------------------
	process (reset, clock)
	begin
		if reset='1'then
			EA <= inicio; -- start by the idle state, called idle
		elsif clock'event and clock='1' then
			EA <= PE;
		end if;
	end process;

	process(start, EA, cont)
	begin
		case EA is
			when inicio =>
				if start='1' then
					PE <= desloca;
				else
					PE <= inicio;
				end if;

			when desloca =>
				PE <= subtrai;

			when subtrai =>
				PE <= esc_quoc;

			when esc_quoc =>
				PE <= restaura;

			when restaura =>
				if cont=size-1 then
					PE <= fim;
				else
					PE <= desloca;
				end if;

			when fim =>
					PE <= inicio;
		end case;
	end process;

end div_serial;